'''
Computes the semantic similarity between two neuron mentions.
The main function is similarity()
Computation is delegated to similarity_inter and similarity_intra
'''

import similarity_inter, similarity_intra
from operator import itemgetter
from itertools import groupby
import re

from sherlok import Sherlok # pip install --upgrade sherlok
s = Sherlok('neuroner')

WEIGHTS = { #TODO: implement weights
    'Layer': 1.0,
    'ProteinProp': 1.0
}

'''
Computes the intra and inter semantic similarity between two neurons
in: n1@str, n2@str: the two neurons to measure similarity
out: (score:float, [([matching_properties], explanation@str)])
'''
def similarity(n1, n2, weights=WEIGHTS):
    assert type(n1) is str and len(n1) > 0, 'n1 cannot be empty'
    assert type(n2) is str and len(n2) > 0, 'n2 cannot be empty'
    n1c = _cleanup(s.annotate(n1).annotations)
    n2c = _cleanup(s.annotate(n2).annotations)
    return similarity2(n1c, n2c)

'''
Cleans up raw Sherlok annotations
in: raw Sherlok output
out: only keep ontologyIds and if none is present concatenate the type and text
'''
def _cleanup(n, orig_neuron_str = None):
    clean = []
    filt_attrib_list = ['Neuron', 'PreNeuron', 'PostNeuron', 'Electrophysiology', 'ProteinTrigger']
    for (begin, end, txt, type_, props) in n:
        if 'ontologyId' in props:
            clean.append((begin, end, props[u'ontologyId']))
        elif type_ not in filt_attrib_list:
            clean.append((begin, end, '{}:{}'.format(type_, txt)))
    clean = sorted(clean, key = lambda tup: tup[0])

    # remove unwanted UNKN_REGION ontology terms if a better ontology term available
    # also defer to NCBI_GENE over HBP_NEUROTRANSMITTER
    pos_list = [c[0] for c in clean]
    new_clean = clean
    for p in pos_list:
        indices = [i for i, x in enumerate(pos_list) if x == p]
        if len(indices) > 1:
            try:
                conflicting_terms = [clean[j] for j in indices]
                for c in conflicting_terms:
                    if 'UNKN' in c[2]:
                        new_clean.remove(c)
                    if 'HBP_NEUROTRANSMITTER' in c[2]:
                        new_clean.remove(c)
            except Exception:
                continue
    clean = new_clean

    # required because sherlok doesn't always return all missing terms
    if orig_neuron_str:
        # do some stuff to return missing terms - really crappy code
        nl = []
        for a in clean:
            for i in range(a[0],a[1]):
                nl.append(i)
            nl.append(i+1)
        unfound_inds = []
        for i in range(1,len(orig_neuron_str)):
            if i in nl:
                continue
            else:
                unfound_inds.append(i)
        data = unfound_inds
        ranges = []
        for k, g in groupby(enumerate(data), lambda (i,x):i-x):
            group = map(itemgetter(1), g)
            ranges.append((group[0], group[-1]))
        for r in ranges:
            l = [r[0], r[1], 'Missing:' + orig_neuron_str[(r[0]-1):(r[1]+1)].strip()]
            clean.append(l)

        clean = sorted(clean, key= lambda tup: tup[0])

    clean = [c[2] for c in clean]
    # filter out neuron triggers because not meaningful
    clean = [c for c in clean if 'NeuronTrigger' not in c]

    return clean

# load all ontologies into one large dictionary
big_onto = similarity_intra.load_ontologies()

def _normalize(onto_list, shorten = False):
    """Convenience function for turning a neuroNER ontology list back into ontology names or acronyms
    """
    normalized_term = []
    for l in onto_list:
        if l in big_onto:
            d = big_onto[l]
            if shorten:
                if 'acronym' in d:
                    normalized_term.append(d['acronym'][0])
                else:
                    normalized_term.append(d['name'])
            else:
                normalized_term.append(d['name'])
        else:
            new_term = re.sub('^\w+:', '', l)
            normalized_term.append(new_term)
    return ' '.join(normalized_term)



'''
Computes the intra and inter semantic similarity between two neurons
in:  n1@[], n2@[]: the two neurons to measure similarity
out: (score:float, [([matching_properties, explanation@str])])
'''
def similarity2(n1, n2, weights=WEIGHTS):
    s_intra = similarity_intra._similarity_intra(n1, n2, weights)
    s_inter = similarity_inter._similarity_inter(n1, n2)
    print('s_intra', s_intra, 's_inter', s_inter)
    return (s_intra[0] + s_inter[0], s_intra[1] + s_inter[1])




'''
Computes the similarity between neuron properties of different property classes.
E.g. compares neurotransmitters (GABA) with proteins (GAD67, and ephys (FS) with proteins (PV)

Similarity rules inter (between) neuron property classes are explicited in the following format:

(['onto_id1', 'onto_id2', ...],     #0: expectations for first neuron
 ['onto_id3'],                      #1: expectations for second neuron
 is_bidirectional,                  #2: can the rule be applied in both directions?
 probability,                       #3: [0 - 1.0]
 'citation',                        #4: a paper supporting this rule
 'explanation'                      #5: explanation or citation snippet
)
'''

_rules = [

# FS <--> PV
(['HBP_EPHYS:0000080', 'HBP_EPHYS_TRIGGER:0000003'], ['NCBI_GENE:19293'], True, 0.9, 'TODO', 'implicit correspondance'),

# Chandelier are FS
(['HBP_MORPHOLOGY:0000007'], ['HBP_EPHYS:0000080', 'HBP_EPHYS_TRIGGER:0000003'], False, 0.9, 'TODO', 'implicit correspondance'),

]


def _similarity_inter(n1, n2):
    (sim_inter, explanations) = (0, [])
    for rule in _rules:
        #print 'rule[0]', rule[0]
        #print 'rule[1]', rule [1]
        sim_this_rule = _sim_rule(rule, n1, n2)
        if rule[2]: # is_bidirectional -> apply rule the other direction
            sim_this_rule += _sim_rule(rule, n2, n1)
        if sim_this_rule > 0:
            sim_inter += sim_this_rule
            explain_text = '{} ({})'.format(rule[5], rule[4])
            explanations.append( (rule[0] + rule[1], explain_text))

    # cap similarity at max 1
    return (min(1, sim_inter), explanations)


# check whether n1 and n2 satisfy that rule
def _sim_rule(rule, n1, n2):
    #print '_sim_rule', rule, n1, n2
    for expectation in rule[0]:
        if expectation not in n1:
            #print('expectation', expectation, 'not in', n1)
            return 0 # n1 does not satisfy expectations
    for expectation in rule[1]:
        if expectation not in n2:
            #print('expectation', expectation, 'not in', n2)
            return 0 # n1 does not satisfy expectations
    return rule[3] # rule probability






'''
Computes the similarity between neuron properties within a property class.
E.g. compares 'layer 1a' with 'layer 1-2' and 'Hypothalamus' with 'Midbrain'
Implemented in XXXXSimilarity classes, e.g. LayerSimilarity, BrainR
'''
import os.path
import oboparser
from similarity_config import cfg
from itertools import chain
import glob

from allensdk.core.mouse_connectivity_cache import MouseConnectivityCache
# requires pip'ing: pip install https://github.com/AllenInstitute/AllenSDK/archive/v0.10.1.tar.gz

onto_root = cfg['onto_root']
assert os.path.exists(onto_root), 'could not find onto_root at ' + onto_root


class LayerSimilarity(object):
    """LayerSimilarity: similarity """
    PREFIX = 'HBP_LAYER'
    def __init__(self):
        self.onto_id2layer_numbers = {}
        # 'HBP_LAYER:0000123' -> 123
        def onto_id2layer_number(o):
            return int(o[10:])
        for o in oboparser.parse(onto_root + 'hbp_layer_ontology.robo'):
            if 'id' in o:
                onto_id = o['id'] # HBP_LAYER:0000001
                assert len(onto_id) == 17, 'invalid onto_id: {} in {}'.format(onto_id, o)
                if onto_id2layer_number(onto_id) < 8: # HBP_LAYER:0000001 to 7
                    # just reslove to layer id
                    self.onto_id2layer_numbers[onto_id] = [onto_id2layer_number(onto_id)]
                elif 'union_of' in o:
                    # put the layer numbers of all 'union_of' ids
                    self.onto_id2layer_numbers[onto_id] = [onto_id2layer_number(u) for u in o['union_of']]
                elif 'is_a' in o: # e.g. L5a is_a L5
                    # to simplify, resolve to L5
                    self.onto_id2layer_numbers[onto_id] = [onto_id2layer_number(o['is_a'][0])]
                else:
                    raise Exception('invalid layer entry: {}'.format(o))

    '''
    returns 1 if both neuron share a similar layer, e.g. 'layer2/3' and 'layer 2A', else 0
    '''
    def similarity(self, n1, n2):
        # HBP_LAYER:0000101 (layer 1-2) --> [1,2]
        def neuron2layer_numbers(neuron):
            layer_numbers = []
            for n in neuron:
                if n.startswith('HBP_LAYER:'):
                    for layer_number in self.onto_id2layer_numbers[n]:
                        layer_numbers.append(layer_number)
            return layer_numbers
        n1_layer_numbers = neuron2layer_numbers(n1)
        n2_layer_numbers = neuron2layer_numbers(n2)
        # on same layer?
        common_layer = set(n1_layer_numbers).intersection(set(n2_layer_numbers))
        if len(common_layer) > 0:
            return (1.0, (['HBP_LAYER:000000{}'.format(common_layer.pop())], 'located on same Layer') )
        else:
            return (0, []) # no layers in both neurons




class BrainRegionSimilarity(object):
    """BrainRegionSimilarity: similarity """
    def __init__(self):
        self.onto_id2parent_regions = {}

        # only find parents up to these major brain regions
        big_reg_acros = ['Isocortex', 'OLF', 'STR', 'PAL', 'TH', 'HY', 'MB', 'PAL', 'MY', 'CB', 'HPF', 'CTXsp']

        mcc = MouseConnectivityCache()
        onto = mcc.get_ontology()
        df = onto.df
        struct_ids = df.id
        for struct_id in struct_ids:
            structure_path_str = onto[struct_id].structure_id_path.item()
            region_path = structure_path_str.split('/')
            parent_list = []
            for r in reversed(region_path):
                if r:
                    parent_list.append(r)
                    tdf = df[df.id == int(r)]
                    acronym = tdf.acronym.item()
                    if acronym in big_reg_acros:
                        break
            self.onto_id2parent_regions[struct_id] = parent_list

    def onto_id2region_id(self, o):
        return int(o[11:])

    # input: n1 and n2, each a list of ontology ids, e.g. ['HBP_EPHYS:0000080', 'HBP_LAYER:0000001', 'NCBI_GENE:19293']
    # returns a tuple with:
    #        1) a score (float) reflecting how similar these neuron are regarding brain regions
    #        2) an array of explanations, each a tuple of the form (['matching_ontology_ids'], 'human readable explanation')
    def similarity(self, n1, n2):
        def neuron2region_ids(neuron):
            region_ids = []
            for n in neuron:
                if n.startswith('ABA_REGION:'):
                    region_id = self.onto_id2region_id(n)
                    region_ids.append(self.onto_id2parent_regions[region_id])
            return list(chain(*region_ids))
        n1_parent_regions = neuron2region_ids(n1)
        n2_parent_regions = neuron2region_ids(n2)
        # print n1_parent_regions
        # print n2_parent_regions

        common_regions = [item for item in n1_parent_regions if item in n2_parent_regions]

        if common_regions:
            if n1_parent_regions[0] == n2_parent_regions[0]:
                return 1.0, (['ABA_REGION:{}'.format(n1_parent_regions[0])], 'exact same brain region')
            elif n1_parent_regions[0] in n2_parent_regions:
                return 1.0, (['ABA_REGION:{}'.format(n1_parent_regions[0])], 'sharing a common brain region')
            elif n2_parent_regions[0] in n1_parent_regions:
                return 1.0, (['ABA_REGION:{}'.format(n2_parent_regions[0])], 'sharing a common brain region')
            elif len(common_regions) > 0:
                return .5, (['ABA_REGION:{}'.format(common_regions.pop())], 'sibling regions')
        else:
            return (0, []) # no regions in both neurons


similarities = [LayerSimilarity(), BrainRegionSimilarity()]


def _similarity_intra(n1, n2, weights):
    # perfect similarity/equality
    if n1 == n2:
        return (1,[])
    else:
        # dispatch to each similarities; aggregate score & explanations
        (sim_intra, explanations) = (0, [])
        for s in similarities:
            (s_sim, s_expl) = s.similarity(n1, n2) #TODO weights
            sim_intra += s_sim; explanations.append(s_expl)
        return (sim_intra, explanations)


def load_ontologies():
    """Loads all of the ontologies into a nice dictionary data structure"""

    # a massive dictionary containing key : dictionary mappings between HBP ontology id's and .obo ontology terms
    big_onto = {}
    mcc = MouseConnectivityCache()
    aba_onto = mcc.get_ontology()

    file_name_list = [f for f in glob.glob(onto_root + "*.robo")]
    file_name_list.extend([f for f in glob.glob(onto_root + "*.obo")])
    for fn in file_name_list:
        for o in oboparser.parse(fn):
            if 'id' in o:
                big_onto[o['id']] = o
    for k in big_onto.keys():
        if 'ABA' in k:
            new_o = big_onto[k]
            aba_id = int(k[11:])
            new_o['acronym'] = [aba_onto[aba_id]['acronym'].item()]
            big_onto[k] = new_o
    return big_onto





# to publish this as a pypi module:
# 1. change (increase) version id in `setup.py`
# 2. python setup.py sdist register upload

from setuptools import setup, find_packages  # Always prefer setuptools over distutils
from codecs import open  # To use a consistent encoding

#with open('README.rst') as f:
#    readme = f.read()

setup(
    name='neuroner',
    version='0.0.2',
    description='named entity recognition for neurons',
    long_description='named entity recognition for neurons',
    url='https://github.com/renaud/neuroNER',
    author='Renaud Richardet, Shreejoy Tripathy',
    author_email='renaud@apache.org, stripathy@chibi.ubc.ca',
    license='Apache License (2.0)',
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 2.7',
         'Operating System :: OS Independent',
         'Environment :: Console',
    ],
    keywords='neuroscience',
    py_modules=['similarity', 'similarity_inter', 'similarity_intra', 'oboparser', 'similarity_config']
)



#!/usr/bin/env python

"""oboparse.py: Some functions to read the OBO format
(http://www.geneontology.org/GO.format.obo-1_2.shtml) and load it
(lazily) in a sequence of Python dicts"""

__author__ = "Stelios Sfakianakis"

from collections import defaultdict
import fileinput
import json


def parse_lines(lines):
    for line in lines:
        i = line.find('!')
        if i >= 0:
            line = line[0:i]
        line = line.strip()
        if line:
            yield line


def read_stanzas(lines):
    def unescape(str):
        return str.strip()
    single_valued = set(['id', 'name', 'is_obsolete', 'obsolete_since',
                        'ontology', 'format-version', 'data-version', 'date',
                        'comment', 'def', 'created_in'])
    stanza = defaultdict(list)
    stanza['@type'] = 'Header'
    for line in lines:
        if line.startswith('['):
            yield stanza
            stanza = defaultdict(list)
            stanza['@type'] = line[1:-1]
        else:
            k, v = line.split(':', 1)
            key = unescape(k)
            val = unescape(v)
            if key in single_valued:
                stanza[key] = val
            else:
                stanza[key].append(val)
    yield stanza

def parse(f):
    return read_stanzas(parse_lines(open(f, 'r')))

if __name__ == '__main__':
    sz = read_stanzas(parse_lines(fileinput.input()))
    for s in sz:
        print json.dumps(s)
